
# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "nvgpu"

@_ods_cext.register_operation(_Dialect)
class DeviceAsyncCopyOp(_ods_ir.OpView):
  r"""
  The `nvgpu.device_async_copy` op initiates an asynchronous copy operation of
  elements from source (global memory) to the destination (shared memory)
  without blocking the thread. The async copy is added to a group.
  
  This op is meant to be used with `nvgpu.device_async_create_group` and
  `nvgpu.device_async_wait` to synchronize copies as explained in those ops
  descriptions.
  
  `bypassL1` attribute is hint to the hardware to bypass the L1 cache during
  async copy, this hint may be ignored by the hardware.
  
  `dstElements` attribute is the total number of elements written to
  destination (shared memory).
  
  `srcElements` argument is the total number of elements read from
  source (global memory).
  
  `srcElements` is an optional argument and when present the op only reads
  `srcElements` number of elements from the source (global memory) and zero fills
  the rest of the elements in the destination (shared memory).
  
  In order to do a copy and wait for the result we need the following
  combination:
  ```
  // copy 1.
  %cp1 = nvgpu.device_async_copy %A[%c0], %B[%c0], 4 :memref<16xf32> to memref<16xf32, 3>
  // copy 2.
  %cp2 = nvgpu.device_async_copy %C[%c0], %D[%c0], 4 : memref<16xf32> to memref<16xf32, 3>
  // group 1 contains copy 1 and copy 2.
  %token1 = nvgpu.device_async_create_group %cp1, %cp2
  // copy 3.
  %cp3 = nvgpu.device_async_copy %E[%c0], %F[%c0], 4 : memref<16xf32> to memref<16xf32, 3>
  // group 2 contains copy 3.
  %token2 = nvgpu.device_async_create_group %cp3
  // after the wait copy 1 and copy 2 are complete.
  nvgpu.device_async_wait %token1
  // after the wait copy 3 is complete.
  nvgpu.device_async_wait %token2
  ```
  
  Example:
  
  ```mlir
  %0 = nvgpu.device_async_copy %src[%c0, %c0], %dst[%c0, %c0, %c0], 4 :
    memref<4x5xf32> to memref<2x7x5xf32, 3>
  ```
  """

  OPERATION_NAME = "nvgpu.device_async_copy"

  _ODS_OPERAND_SEGMENTS = [1,-1,1,-1,0,]

  _ODS_REGIONS = (0, True)

  def __init__(self, dst: _ods_ir.Value[_ods_ir.MemRefType], dstIndices: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], src: _ods_ir.Value[_ods_ir.MemRefType], srcIndices: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], dstElements: _Union[int, _ods_ir.IntegerAttr], *, srcElements: _Optional[_ods_ir.Value[_ods_ir.IndexType]] = None, bypassL1: _Optional[bool] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(dst)
    operands.append(_get_op_results_or_values(dstIndices))
    operands.append(src)
    operands.append(_get_op_results_or_values(srcIndices))
    operands.append(srcElements)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["dstElements"] = (dstElements if (
    isinstance(dstElements, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('IndexAttr')) else
      _ods_ir.AttrBuilder.get('IndexAttr')(dstElements, context=_ods_context))
    if bool(bypassL1): attributes["bypassL1"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def dst(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range[0]

  @builtins.property
  def dstIndices(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def src(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 2)
    return operand_range[0]

  @builtins.property
  def srcIndices(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 3)
    return operand_range

  @builtins.property
  def srcElements(self) -> _Optional[_ods_ir.Value[_ods_ir.IndexType]]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 4)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def dstElements(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["dstElements"]

  @dstElements.setter
  def dstElements(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["dstElements"] = value

  @builtins.property
  def bypassL1(self) -> bool:
    return "bypassL1" in self.operation.attributes

  @bypassL1.setter
  def bypassL1(self, value):
    if bool(value):
      self.operation.attributes["bypassL1"] = _ods_ir.UnitAttr.get()
    elif "bypassL1" in self.operation.attributes:
      del self.operation.attributes["bypassL1"]

  @bypassL1.deleter
  def bypassL1(self):
    del self.operation.attributes["bypassL1"]

  @builtins.property
  def asyncToken(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(DeviceAsyncCopyOp)
class DeviceAsyncCopyOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.device_async_copy"

  @builtins.property
  def dst(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range[0]

  @builtins.property
  def dstIndices(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def src(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 2)
    return operand_range[0]

  @builtins.property
  def srcIndices(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 3)
    return operand_range

  @builtins.property
  def srcElements(self) -> _Optional[_ods_ir.Value[_ods_ir.IndexType]]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 4)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def dstElements(self) -> _ods_ir.IntegerAttr:
    return self.attributes["dstElements"]

  @builtins.property
  def bypassL1(self) -> bool:
    return "bypassL1" in self.attributes

def device_async_copy(dst: _ods_ir.Value[_ods_ir.MemRefType], dst_indices: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], src: _ods_ir.Value[_ods_ir.MemRefType], src_indices: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], dst_elements: _Union[int, _ods_ir.IntegerAttr], *, src_elements: _Optional[_ods_ir.Value[_ods_ir.IndexType]] = None, bypass_l1: _Optional[bool] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return DeviceAsyncCopyOp(dst=dst, dstIndices=dst_indices, src=src, srcIndices=src_indices, dstElements=dst_elements, srcElements=src_elements, bypassL1=bypass_l1, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class DeviceAsyncCreateGroupOp(_ods_ir.OpView):
  r"""
    The `nvgpu.device_async_create_group` op creates a group of memory accesses
    containing all the pending `device_async_copy` operations associated with
    argument tokens. Each token can only be part of one group.
  
    It returns a token that can be use to wait until the group fully completes.
  
    This is meant to be used with `nvgpu.device_async_wait` to synchronize copies
    as explained in those ops descriptions.
  
    Groups are executed in the order they are created.
  
    Example:
  
    ```mlir
    %0 = nvgpu.device_async_create_group
  ```
  """

  OPERATION_NAME = "nvgpu.device_async_create_group"

  _ODS_REGIONS = (0, True)

  def __init__(self, inputTokens: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(inputTokens))
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def inputTokens(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def asyncToken(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(DeviceAsyncCreateGroupOp)
class DeviceAsyncCreateGroupOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.device_async_create_group"

  @builtins.property
  def inputTokens(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def device_async_create_group(input_tokens: _Sequence[_ods_ir.Value], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return DeviceAsyncCreateGroupOp(inputTokens=input_tokens, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class DeviceAsyncWaitOp(_ods_ir.OpView):
  r"""
  The `nvgpu.device_async_wait` op will block the execution thread until the group
  associated with the source token is fully completed.
  
  The optional `$numGroups` attribute gives an upper bound of the number of
  groups uncompleted when the wait can unblock the thread. For example,  if
  16 async groups are pushe and `$numGroups` is set to 12, then the thread
  will unblock when 12 groups or fewer are in flight (4 groups have
  completed).
  
  Example:
  
  ```mlir
  nvgpu.device_async_wait %0
  ```
  """

  OPERATION_NAME = "nvgpu.device_async_wait"

  _ODS_REGIONS = (0, True)

  def __init__(self, asyncDependencies: _ods_ir.Value, *, numGroups: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(asyncDependencies)
    _ods_context = _ods_get_default_loc_context(loc)
    if numGroups is not None: attributes["numGroups"] = (numGroups if (
        isinstance(numGroups, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I32Attr')) else
          _ods_ir.AttrBuilder.get('I32Attr')(numGroups, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def asyncDependencies(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def numGroups(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "numGroups" not in self.operation.attributes:
      return None
    return self.operation.attributes["numGroups"]

  @numGroups.setter
  def numGroups(self, value: _Optional[_ods_ir.IntegerAttr]):
    if value is not None:
      self.operation.attributes["numGroups"] = value
    elif "numGroups" in self.operation.attributes:
      del self.operation.attributes["numGroups"]

  @numGroups.deleter
  def numGroups(self):
    del self.operation.attributes["numGroups"]

@_ods_cext.register_op_adaptor(DeviceAsyncWaitOp)
class DeviceAsyncWaitOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.device_async_wait"

  @builtins.property
  def asyncDependencies(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def numGroups(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "numGroups" not in self.attributes:
      return None
    return self.attributes["numGroups"]

def device_async_wait(async_dependencies: _ods_ir.Value, *, num_groups: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> DeviceAsyncWaitOp:
  return DeviceAsyncWaitOp(asyncDependencies=async_dependencies, numGroups=num_groups, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class LdMatrixOp(_ods_ir.OpView):
  r"""
  The `nvgpu.ldmatrix` op represents loading a matrix fragment from
  memory to registers. The source and result type must be compatible
  with lowering to the `nvvm.ldmatrix` instruction. This op represents
  the distributed version of a `vector.transfer_read` as an intermediate
  step between lowering from `vector.transfer_read` to `nvvm.ldmatrix`.
  
  This operation is meant to follow the semantic of described here:
  https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#warp-level-matrix-instructions-ldmatrix
  
  Example:
  ```mlir
  %0 = nvgpu.ldmatrix %sm[%c0, %c0] {numTiles = 4 : i32, transpose = false} :
    memref<?x?xf16, 3> -> vector<4x2xf16>
  ```
  """

  OPERATION_NAME = "nvgpu.ldmatrix"

  _ODS_REGIONS = (0, True)

  def __init__(self, res: _ods_ir.Type, srcMemref: _ods_ir.Value[_ods_ir.MemRefType], indices: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], transpose: _Union[bool, _ods_ir.BoolAttr], numTiles: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(srcMemref)
    operands.extend(_get_op_results_or_values(indices))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["transpose"] = (transpose if (
    isinstance(transpose, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('BoolAttr')) else
      _ods_ir.AttrBuilder.get('BoolAttr')(transpose, context=_ods_context))
    attributes["numTiles"] = (numTiles if (
    isinstance(numTiles, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I32Attr')) else
      _ods_ir.AttrBuilder.get('I32Attr')(numTiles, context=_ods_context))
    results = []
    results.append(res)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def srcMemref(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    return self.operation.operands[0]

  @builtins.property
  def indices(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
    return self.operation.operands[1:1 + _ods_variadic_group_length]

  @builtins.property
  def transpose(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["transpose"]

  @transpose.setter
  def transpose(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["transpose"] = value

  @builtins.property
  def numTiles(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["numTiles"]

  @numTiles.setter
  def numTiles(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["numTiles"] = value

  @builtins.property
  def res(self) -> _ods_ir.OpResult[_ods_ir.VectorType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(LdMatrixOp)
class LdMatrixOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.ldmatrix"

  @builtins.property
  def srcMemref(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    return self.operands[0]

  @builtins.property
  def indices(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    _ods_variadic_group_length = len(self.operands) - 2 + 1
    return self.operands[1:1 + _ods_variadic_group_length]

  @builtins.property
  def transpose(self) -> _ods_ir.BoolAttr:
    return self.attributes["transpose"]

  @builtins.property
  def numTiles(self) -> _ods_ir.IntegerAttr:
    return self.attributes["numTiles"]

def ldmatrix(res: _ods_ir.Type, src_memref: _ods_ir.Value[_ods_ir.MemRefType], indices: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], transpose: _Union[bool, _ods_ir.BoolAttr], num_tiles: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.VectorType]:
  return LdMatrixOp(res=res, srcMemref=src_memref, indices=indices, transpose=transpose, numTiles=num_tiles, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MBarrierArriveExpectTxOp(_ods_ir.OpView):
  r"""
  A thread executing the Op performs an expect-tx operation on the mbarrier 
  object at the location specified by the address operand $barrier. The 
  expect-tx operation, with an $txcount argument, increases the tx-count of 
  an mbarrier object by the value specified by $txcount. This makes the 
  current phase of the mbarrier object to expect and track the completion of 
  additional asynchronous transactions.
  
  The `$txCount` specifies the number of element to the expect-tx operation.
  
  Example:
  ```mlir
    nvgpu.mbarrier.arrive.expect_tx %barrier, %ic0 : !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>>
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.arrive.expect_tx"

  _ODS_REGIONS = (0, True)

  def __init__(self, barriers: _ods_ir.Value, txcount: _ods_ir.Value[_ods_ir.IndexType], mbarId: _ods_ir.Value[_ods_ir.IndexType], *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(barriers)
    operands.append(txcount)
    operands.append(mbarId)
    if predicate is not None: operands.append(predicate)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def txcount(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[2]

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    return None if len(self.operation.operands) < 4 else self.operation.operands[3]

@_ods_cext.register_op_adaptor(MBarrierArriveExpectTxOp)
class MBarrierArriveExpectTxOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.arrive.expect_tx"

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def txcount(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[2]

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    return None if len(self.operands) < 4 else self.operands[3]

def mbarrier_arrive_expect_tx(barriers: _ods_ir.Value, txcount: _ods_ir.Value[_ods_ir.IndexType], mbar_id: _ods_ir.Value[_ods_ir.IndexType], *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> MBarrierArriveExpectTxOp:
  return MBarrierArriveExpectTxOp(barriers=barriers, txcount=txcount, mbarId=mbar_id, predicate=predicate, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class MBarrierArriveNoCompleteOp(_ods_ir.OpView):
  r"""
  The Op performs arrive-on operation on the `mbarrier` object and returns a 
  `nvgpu.mbarrier.token`.
  
  The Op does not cause the `nvgpu.mbarrier` to complete its current phase.
  
  Example:
  ```mlir
    %token = nvgpu.mbarrier.arrive.noComplete %barrier, %count : !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>> -> !nvgpu.mbarrier.token
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.arrive.nocomplete"

  _ODS_REGIONS = (0, True)

  def __init__(self, barriers: _ods_ir.Value, mbarId: _ods_ir.Value[_ods_ir.IndexType], count: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(barriers)
    operands.append(mbarId)
    operands.append(count)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def count(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[2]

  @builtins.property
  def token(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MBarrierArriveNoCompleteOp)
class MBarrierArriveNoCompleteOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.arrive.nocomplete"

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

  @builtins.property
  def count(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[2]

def mbarrier_arrive_nocomplete(barriers: _ods_ir.Value, mbar_id: _ods_ir.Value[_ods_ir.IndexType], count: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return MBarrierArriveNoCompleteOp(barriers=barriers, mbarId=mbar_id, count=count, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MBarrierArriveOp(_ods_ir.OpView):
  r"""
  The Op performs arrive-on operation on the `mbarrier` object and returns a 
  `nvgpu.mbarrier.token`.
  
  For more information, see
  https://docs.nvidia.com/cuda/parallel-thread-execution/#arrive-on-operation-on-mbarrier-object
  
  Example:
  ```mlir
    %token = nvgpu.mbarrier.arrive %barrier : !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>> -> !nvgpu.mbarrier.token
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.arrive"

  _ODS_REGIONS = (0, True)

  def __init__(self, barriers: _ods_ir.Value, mbarId: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(barriers)
    operands.append(mbarId)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def token(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MBarrierArriveOp)
class MBarrierArriveOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.arrive"

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def mbarrier_arrive(barriers: _ods_ir.Value, mbar_id: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return MBarrierArriveOp(barriers=barriers, mbarId=mbar_id, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MBarrierCreateOp(_ods_ir.OpView):
  r"""
  The Op generates one or more `mbarrier` object, which is a barrier created in 
  shared memory and supports various synchronization behaviors for threads.
  
  The `mbarrier` object has the following type and alignment requirements:
    Type: .b64, Alignment: 8, Memory space: .shared
  
  Example:
  ```mlir
    %barrier = nvgpu.mbarrier.create -> !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>>
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.create"

  _ODS_REGIONS = (0, True)

  def __init__(self, barriers: _ods_ir.Type, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(barriers)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MBarrierCreateOp)
class MBarrierCreateOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.create"

def mbarrier_create(barriers: _ods_ir.Type, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return MBarrierCreateOp(barriers=barriers, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MBarrierGetOp(_ods_ir.OpView):
  r"""
  The `nvgpu.mbarrier.get` operation retrieves a pointer to a specific 
  `mbarrier` object from a group of barriers created by the `nvgpu.mbarrier.create` operation.
  
  Example:
  ```mlir
    %mbars = nvgpu.mbarrier.create -> !nvgpu.mbarrier.group<memorySpace = #gpu.address_space<workgroup>, num_barriers = 10>
    %mbar_pointer = nvgpu.mbarrier.get %mbars[%c2] : !nvgpu.mbarrier.group<memorySpace = #gpu.address_space<workgroup>>
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.get"

  _ODS_REGIONS = (0, True)

  def __init__(self, mbarrierPointer: _ods_ir.Type, barriers: _ods_ir.Value, mbarId: _ods_ir.Value[_ods_ir.IndexType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(barriers)
    operands.append(mbarId)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(mbarrierPointer)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def mbarrierPointer(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MBarrierGetOp)
class MBarrierGetOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.get"

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def mbarrier_get(mbarrier_pointer: _ods_ir.Type, barriers: _ods_ir.Value, mbar_id: _ods_ir.Value[_ods_ir.IndexType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return MBarrierGetOp(mbarrierPointer=mbarrier_pointer, barriers=barriers, mbarId=mbar_id, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MBarrierInitOp(_ods_ir.OpView):
  r"""
  The Op initializes the `mbarrier` object with the given number of threads.
  
  Example:
  ```mlir
    %num_threads = gpu.block_dim x
    %barrier = nvgpu.mbarrier.create -> !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>>
    nvgpu.mbarrier.init %barrier, %num_threads : !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>>
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.init"

  _ODS_REGIONS = (0, True)

  def __init__(self, barriers: _ods_ir.Value, count: _ods_ir.Value[_ods_ir.IndexType], mbarId: _ods_ir.Value[_ods_ir.IndexType], *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(barriers)
    operands.append(count)
    operands.append(mbarId)
    if predicate is not None: operands.append(predicate)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def count(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[2]

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    return None if len(self.operation.operands) < 4 else self.operation.operands[3]

@_ods_cext.register_op_adaptor(MBarrierInitOp)
class MBarrierInitOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.init"

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def count(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[2]

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    return None if len(self.operands) < 4 else self.operands[3]

def mbarrier_init(barriers: _ods_ir.Value, count: _ods_ir.Value[_ods_ir.IndexType], mbar_id: _ods_ir.Value[_ods_ir.IndexType], *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> MBarrierInitOp:
  return MBarrierInitOp(barriers=barriers, count=count, mbarId=mbar_id, predicate=predicate, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class MBarrierTestWaitOp(_ods_ir.OpView):
  r"""
  Checks whether the mbarrier object has completed the phase. It is is a 
  non-blocking instruction which tests for the completion of the phase.
  
  Example:
  ```mlir
    %isComplete = nvgpu.mbarrier.test.wait %barrier, %token : !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>>, !nvgpu.mbarrier.token
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.test.wait"

  _ODS_REGIONS = (0, True)

  def __init__(self, barriers: _ods_ir.Value, token: _ods_ir.Value, mbarId: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(barriers)
    operands.append(token)
    operands.append(mbarId)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def token(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[2]

  @builtins.property
  def waitComplete(self) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MBarrierTestWaitOp)
class MBarrierTestWaitOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.test.wait"

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def token(self) -> _ods_ir.Value:
    return self.operands[1]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[2]

def mbarrier_test_wait(barriers: _ods_ir.Value, token: _ods_ir.Value, mbar_id: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
  return MBarrierTestWaitOp(barriers=barriers, token=token, mbarId=mbar_id, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MBarrierTryWaitParityOp(_ods_ir.OpView):
  r"""
  Checks whether the mbarrier object has completed the phase. It is is a 
  potentially blocking instruction which tests for the completion of the 
  phase. Suspended thread resumes execution when the specified phase completes 
  OR before the phase completes following a system-dependent time limit. 
  
  The `$phaseParity` specifies either even phase (0) or odd phase (1) to 
  wait.
  
  Example:
  ```mlir
    nvgpu.mbarrier.try_wait.parity %barrier, %phaseParity, %ticks : !nvgpu.mbarrier.barrier<memorySpace = #gpu.address_space<workgroup>>
  ```
  """

  OPERATION_NAME = "nvgpu.mbarrier.try_wait.parity"

  _ODS_REGIONS = (0, True)

  def __init__(self, barriers: _ods_ir.Value, phaseParity: _ods_ir.Value[_ods_ir.IntegerType], ticks: _ods_ir.Value[_ods_ir.IndexType], mbarId: _ods_ir.Value[_ods_ir.IndexType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(barriers)
    operands.append(phaseParity)
    operands.append(ticks)
    operands.append(mbarId)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def phaseParity(self) -> _ods_ir.Value[_ods_ir.IntegerType]:
    return self.operation.operands[1]

  @builtins.property
  def ticks(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[2]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[3]

@_ods_cext.register_op_adaptor(MBarrierTryWaitParityOp)
class MBarrierTryWaitParityOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mbarrier.try_wait.parity"

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def phaseParity(self) -> _ods_ir.Value[_ods_ir.IntegerType]:
    return self.operands[1]

  @builtins.property
  def ticks(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[2]

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[3]

def mbarrier_try_wait_parity(barriers: _ods_ir.Value, phase_parity: _ods_ir.Value[_ods_ir.IntegerType], ticks: _ods_ir.Value[_ods_ir.IndexType], mbar_id: _ods_ir.Value[_ods_ir.IndexType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> MBarrierTryWaitParityOp:
  return MBarrierTryWaitParityOp(barriers=barriers, phaseParity=phase_parity, ticks=ticks, mbarId=mbar_id, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class MmaSparseSyncOp(_ods_ir.OpView):
  r"""
  The `nvgu.mma.sp.sync` operation performs a warp-distributed MMA operation
  where operand A is "structured sparse". In this case, the `matrixA` operand
  represents the (warp-distributed) non-zero values of operand A, and the
  `sparse_metadata` operand provides the indices.
  
  The full description of the sparsity storage format and distribution scheme is
  described in the PTX docs. This operation is meant to follow the semantic
  described in the PTX documentation here:
  https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#warp-level-matrix-instructions-for-sparse-mma
  
  The way the indices are distributed among the threads in a warp is controlled
  by the optional `sparsity_selector` operand, which is `0` by default. For
  more information, please consult the PTX documentation linked above.
  
  Example (targetingthe f16 16x8x32 `mma.sp` PTX instruction):
  
  ```mlir
  nvgpu.mma.sp.sync (%a, %b, %c) metadata (%meta) {mmaShape = [16, 8, 32]} :
    (vector<4x2xf16>, vector<2x2xf16>, vector<2x2xf16>) -> vector<2x2xf16>
  ```
  """

  OPERATION_NAME = "nvgpu.mma.sp.sync"

  _ODS_REGIONS = (0, True)

  def __init__(self, res: _ods_ir.Type, matrixA: _ods_ir.Value[_ods_ir.VectorType], matrixB: _ods_ir.Value[_ods_ir.VectorType], matrixC: _ods_ir.Value[_ods_ir.VectorType], sparseMetadata: _ods_ir.Value[_ods_ir.VectorType], mmaShape: _Union[_Sequence[int], _ods_ir.ArrayAttr], *, sparsitySelector: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, tf32Enabled: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(matrixA)
    operands.append(matrixB)
    operands.append(matrixC)
    operands.append(sparseMetadata)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["mmaShape"] = (mmaShape if (
    isinstance(mmaShape, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64ArrayAttr')) else
      _ods_ir.AttrBuilder.get('I64ArrayAttr')(mmaShape, context=_ods_context))
    if sparsitySelector is not None: attributes["sparsitySelector"] = (sparsitySelector if (
        isinstance(sparsitySelector, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I32Attr')) else
          _ods_ir.AttrBuilder.get('I32Attr')(sparsitySelector, context=_ods_context))
    if bool(tf32Enabled): attributes["tf32Enabled"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.append(res)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def matrixA(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[0]

  @builtins.property
  def matrixB(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[1]

  @builtins.property
  def matrixC(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[2]

  @builtins.property
  def sparseMetadata(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[3]

  @builtins.property
  def mmaShape(self) -> _ods_ir.ArrayAttr:
    return self.operation.attributes["mmaShape"]

  @mmaShape.setter
  def mmaShape(self, value: _ods_ir.ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["mmaShape"] = value

  @builtins.property
  def sparsitySelector(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["sparsitySelector"]

  @sparsitySelector.setter
  def sparsitySelector(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sparsitySelector"] = value

  @builtins.property
  def tf32Enabled(self) -> bool:
    return "tf32Enabled" in self.operation.attributes

  @tf32Enabled.setter
  def tf32Enabled(self, value):
    if bool(value):
      self.operation.attributes["tf32Enabled"] = _ods_ir.UnitAttr.get()
    elif "tf32Enabled" in self.operation.attributes:
      del self.operation.attributes["tf32Enabled"]

  @tf32Enabled.deleter
  def tf32Enabled(self):
    del self.operation.attributes["tf32Enabled"]

  @builtins.property
  def res(self) -> _ods_ir.OpResult[_ods_ir.VectorType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MmaSparseSyncOp)
class MmaSparseSyncOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mma.sp.sync"

  @builtins.property
  def matrixA(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[0]

  @builtins.property
  def matrixB(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[1]

  @builtins.property
  def matrixC(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[2]

  @builtins.property
  def sparseMetadata(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[3]

  @builtins.property
  def mmaShape(self) -> _ods_ir.ArrayAttr:
    return self.attributes["mmaShape"]

  @builtins.property
  def sparsitySelector(self) -> _ods_ir.IntegerAttr:
    return self.attributes["sparsitySelector"]

  @builtins.property
  def tf32Enabled(self) -> bool:
    return "tf32Enabled" in self.attributes

def mma_sp_sync(res: _ods_ir.Type, matrix_a: _ods_ir.Value[_ods_ir.VectorType], matrix_b: _ods_ir.Value[_ods_ir.VectorType], matrix_c: _ods_ir.Value[_ods_ir.VectorType], sparse_metadata: _ods_ir.Value[_ods_ir.VectorType], mma_shape: _Union[_Sequence[int], _ods_ir.ArrayAttr], *, sparsity_selector: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, tf32_enabled: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.VectorType]:
  return MmaSparseSyncOp(res=res, matrixA=matrix_a, matrixB=matrix_b, matrixC=matrix_c, sparseMetadata=sparse_metadata, mmaShape=mma_shape, sparsitySelector=sparsity_selector, tf32Enabled=tf32_enabled, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MmaSyncOp(_ods_ir.OpView):
  r"""
  The `nvgpu.mma.sync` op represents the warp-level matrix-multiply-and-
  accumulate (mma) operation that is compatible with `nvvm.mma.sync`.
  The operands and results vector sizes are thread-level onwership to
  the warp-level mma operation shape. `mmaShape` attribute holds the
  warp-level matrix-multiply shape.
  
  The `nvgpu.mma.sync` op serves as an intermediate point between lowering from
  `vector.contract` to `nvvm.mma.sync`.
  
  This operation is meant to follow the semantic of described here:
    https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#warp-level-matrix-instructions-mma
  
  Example:
  
  ```mlir
  %res = nvgpu.mma.sync (%matrixA, %matrixB, %matrixC) {mmaShape = [16, 8, 16]} :
      (vector<4x2xf16>, vector<2x2xf16>, vector<2x2xf32>) -> vector<2x2xf32>
  ```
  """

  OPERATION_NAME = "nvgpu.mma.sync"

  _ODS_REGIONS = (0, True)

  def __init__(self, res: _ods_ir.Type, matrixA: _ods_ir.Value[_ods_ir.VectorType], matrixB: _ods_ir.Value[_ods_ir.VectorType], matrixC: _ods_ir.Value[_ods_ir.VectorType], mmaShape: _Union[_Sequence[int], _ods_ir.ArrayAttr], *, tf32Enabled: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(matrixA)
    operands.append(matrixB)
    operands.append(matrixC)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["mmaShape"] = (mmaShape if (
    isinstance(mmaShape, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64ArrayAttr')) else
      _ods_ir.AttrBuilder.get('I64ArrayAttr')(mmaShape, context=_ods_context))
    if bool(tf32Enabled): attributes["tf32Enabled"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.append(res)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def matrixA(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[0]

  @builtins.property
  def matrixB(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[1]

  @builtins.property
  def matrixC(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[2]

  @builtins.property
  def mmaShape(self) -> _ods_ir.ArrayAttr:
    return self.operation.attributes["mmaShape"]

  @mmaShape.setter
  def mmaShape(self, value: _ods_ir.ArrayAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["mmaShape"] = value

  @builtins.property
  def tf32Enabled(self) -> bool:
    return "tf32Enabled" in self.operation.attributes

  @tf32Enabled.setter
  def tf32Enabled(self, value):
    if bool(value):
      self.operation.attributes["tf32Enabled"] = _ods_ir.UnitAttr.get()
    elif "tf32Enabled" in self.operation.attributes:
      del self.operation.attributes["tf32Enabled"]

  @tf32Enabled.deleter
  def tf32Enabled(self):
    del self.operation.attributes["tf32Enabled"]

  @builtins.property
  def res(self) -> _ods_ir.OpResult[_ods_ir.VectorType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MmaSyncOp)
class MmaSyncOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.mma.sync"

  @builtins.property
  def matrixA(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[0]

  @builtins.property
  def matrixB(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[1]

  @builtins.property
  def matrixC(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[2]

  @builtins.property
  def mmaShape(self) -> _ods_ir.ArrayAttr:
    return self.attributes["mmaShape"]

  @builtins.property
  def tf32Enabled(self) -> bool:
    return "tf32Enabled" in self.attributes

def mma_sync(res: _ods_ir.Type, matrix_a: _ods_ir.Value[_ods_ir.VectorType], matrix_b: _ods_ir.Value[_ods_ir.VectorType], matrix_c: _ods_ir.Value[_ods_ir.VectorType], mma_shape: _Union[_Sequence[int], _ods_ir.ArrayAttr], *, tf32_enabled: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.VectorType]:
  return MmaSyncOp(res=res, matrixA=matrix_a, matrixB=matrix_b, matrixC=matrix_c, mmaShape=mma_shape, tf32Enabled=tf32_enabled, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RcpOp(_ods_ir.OpView):
  r"""
  Reciprocal calculation for `vector` types using `nvvm.rcp` OPs.
  
  Currently, only the `approx` rounding mode and `ftz` are supported, and only for the `f32` type.
  
  The input and output must be of the same vector type and shape.
  """

  OPERATION_NAME = "nvgpu.rcp"

  _ODS_REGIONS = (0, True)

  def __init__(self, in_: _ods_ir.Value[_ods_ir.VectorType], *, rounding: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, ftz: _Optional[bool] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(in_)
    _ods_context = _ods_get_default_loc_context(loc)
    if rounding is not None: attributes["rounding"] = (rounding if (
        isinstance(rounding, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('nvgpu.RcpRoundingModeAttr')) else
          _ods_ir.AttrBuilder.get('nvgpu.RcpRoundingModeAttr')(rounding, context=_ods_context))
    if bool(ftz): attributes["ftz"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    if results is None: results = [operands[0].type] * 1
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def in_(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operation.operands[0]

  @builtins.property
  def rounding(self) -> _ods_ir.Attribute:
    return self.operation.attributes["rounding"]

  @rounding.setter
  def rounding(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["rounding"] = value

  @builtins.property
  def ftz(self) -> bool:
    return "ftz" in self.operation.attributes

  @ftz.setter
  def ftz(self, value):
    if bool(value):
      self.operation.attributes["ftz"] = _ods_ir.UnitAttr.get()
    elif "ftz" in self.operation.attributes:
      del self.operation.attributes["ftz"]

  @ftz.deleter
  def ftz(self):
    del self.operation.attributes["ftz"]

  @builtins.property
  def out(self) -> _ods_ir.OpResult[_ods_ir.VectorType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RcpOp)
class RcpOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.rcp"

  @builtins.property
  def in_(self) -> _ods_ir.Value[_ods_ir.VectorType]:
    return self.operands[0]

  @builtins.property
  def rounding(self) -> _ods_ir.Attribute:
    return self.attributes["rounding"]

  @builtins.property
  def ftz(self) -> bool:
    return "ftz" in self.attributes

def rcp(in_: _ods_ir.Value[_ods_ir.VectorType], *, rounding: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, ftz: _Optional[bool] = None, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.VectorType]:
  return RcpOp(in_=in_, rounding=rounding, ftz=ftz, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class TmaAsyncLoadOp(_ods_ir.OpView):
  r"""
  The Op loads a tile memory region from global memory to shared memory by 
  Tensor Memory Access (TMA).
  
  `$tensorMapDescriptor` is tensor map descriptor which has information about
  tile shape. The descriptor is created by `nvgpu.tma.create.descriptor`
  
  The Op uses `$barrier` mbarrier based completion mechanism. 
  """

  OPERATION_NAME = "nvgpu.tma.async.load"

  _ODS_OPERAND_SEGMENTS = [1,1,1,-1,1,0,0,]

  _ODS_REGIONS = (0, True)

  def __init__(self, dst: _ods_ir.Value[_ods_ir.MemRefType], barriers: _ods_ir.Value, tensorMapDescriptor: _ods_ir.Value, coordinates: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], mbarId: _ods_ir.Value[_ods_ir.IndexType], *, multicastMask: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(dst)
    operands.append(barriers)
    operands.append(tensorMapDescriptor)
    operands.append(_get_op_results_or_values(coordinates))
    operands.append(mbarId)
    operands.append(multicastMask)
    operands.append(predicate)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def dst(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range[0]

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range[0]

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 2)
    return operand_range[0]

  @builtins.property
  def coordinates(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 3)
    return operand_range

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 4)
    return operand_range[0]

  @builtins.property
  def multicastMask(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 5)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 6)
    return operand_range[0] if len(operand_range) > 0 else None

@_ods_cext.register_op_adaptor(TmaAsyncLoadOp)
class TmaAsyncLoadOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.tma.async.load"

  @builtins.property
  def dst(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range[0]

  @builtins.property
  def barriers(self) -> _ods_ir.Value:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range[0]

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 2)
    return operand_range[0]

  @builtins.property
  def coordinates(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 3)
    return operand_range

  @builtins.property
  def mbarId(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 4)
    return operand_range[0]

  @builtins.property
  def multicastMask(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 5)
    return operand_range[0] if len(operand_range) > 0 else None

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 6)
    return operand_range[0] if len(operand_range) > 0 else None

def tma_async_load(dst: _ods_ir.Value[_ods_ir.MemRefType], barriers: _ods_ir.Value, tensor_map_descriptor: _ods_ir.Value, coordinates: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], mbar_id: _ods_ir.Value[_ods_ir.IndexType], *, multicast_mask: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> TmaAsyncLoadOp:
  return TmaAsyncLoadOp(dst=dst, barriers=barriers, tensorMapDescriptor=tensor_map_descriptor, coordinates=coordinates, mbarId=mbar_id, multicastMask=multicast_mask, predicate=predicate, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class TmaAsyncStoreOp(_ods_ir.OpView):
  r"""
  The Op store a tile memory region from global memory to shared memory by 
  Tensor Memory Access (TMA).
  
  `$tensorMapDescriptor` is tensor map descriptor which has information about
  tile shape. The descriptor is created by `nvgpu.tma.create.descriptor`
  """

  OPERATION_NAME = "nvgpu.tma.async.store"

  _ODS_OPERAND_SEGMENTS = [1,1,-1,0,]

  _ODS_REGIONS = (0, True)

  def __init__(self, src: _ods_ir.Value[_ods_ir.MemRefType], tensorMapDescriptor: _ods_ir.Value, coordinates: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(src)
    operands.append(tensorMapDescriptor)
    operands.append(_get_op_results_or_values(coordinates))
    operands.append(predicate)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def src(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range[0]

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range[0]

  @builtins.property
  def coordinates(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 2)
    return operand_range

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 3)
    return operand_range[0] if len(operand_range) > 0 else None

@_ods_cext.register_op_adaptor(TmaAsyncStoreOp)
class TmaAsyncStoreOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.tma.async.store"

  @builtins.property
  def src(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range[0]

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range[0]

  @builtins.property
  def coordinates(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 2)
    return operand_range

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 3)
    return operand_range[0] if len(operand_range) > 0 else None

def tma_async_store(src: _ods_ir.Value[_ods_ir.MemRefType], tensor_map_descriptor: _ods_ir.Value, coordinates: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> TmaAsyncStoreOp:
  return TmaAsyncStoreOp(src=src, tensorMapDescriptor=tensor_map_descriptor, coordinates=coordinates, predicate=predicate, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class TmaCreateDescriptorOp(_ods_ir.OpView):
  r"""
  The Op creates a tensor map descriptor object representing tiled memory 
  region. To do that it calls CUDA Driver's `cuTensorMapEncodeTiled`. The 
  descriptor is used by Tensor Memory Access (TMA).
  
  The `tensor` is the source tensor to be tiled. 
  
  The `boxDimensions` is the size of the tiled memory region in each dimension.
  
  For more information see below:
  https://docs.nvidia.com/cuda/cuda-driver-api/group__CUDA__TENSOR__MEMORY.html
  """

  OPERATION_NAME = "nvgpu.tma.create.descriptor"

  _ODS_REGIONS = (0, True)

  def __init__(self, tensorMap: _ods_ir.Type, tensor: _ods_ir.Value[_ods_ir.UnrankedMemRefType], boxDimensions: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(tensor)
    operands.extend(_get_op_results_or_values(boxDimensions))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(tensorMap)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def tensor(self) -> _ods_ir.Value[_ods_ir.UnrankedMemRefType]:
    return self.operation.operands[0]

  @builtins.property
  def boxDimensions(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
    return self.operation.operands[1:1 + _ods_variadic_group_length]

  @builtins.property
  def tensorMap(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(TmaCreateDescriptorOp)
class TmaCreateDescriptorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.tma.create.descriptor"

  @builtins.property
  def tensor(self) -> _ods_ir.Value[_ods_ir.UnrankedMemRefType]:
    return self.operands[0]

  @builtins.property
  def boxDimensions(self) -> _ods_ir.OpOperandList[_ods_ir.IndexType]:
    _ods_variadic_group_length = len(self.operands) - 2 + 1
    return self.operands[1:1 + _ods_variadic_group_length]

def tma_create_descriptor(tensor_map: _ods_ir.Type, tensor: _ods_ir.Value[_ods_ir.UnrankedMemRefType], box_dimensions: _Sequence[_ods_ir.Value[_ods_ir.IndexType]], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return TmaCreateDescriptorOp(tensorMap=tensor_map, tensor=tensor, boxDimensions=box_dimensions, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class TmaFenceOp(_ods_ir.OpView):
  r"""
  The Op fences the given `$tmaDescriptor`. This is necessary if the tensor map
  descriptor was modified from the host using cudaMemcpy. In this case, the
  kernel needs a fence after which it is safe to use `tensor.map`.
  """

  OPERATION_NAME = "nvgpu.tma.fence.descriptor"

  _ODS_REGIONS = (0, True)

  def __init__(self, tensorMapDescriptor: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(tensorMapDescriptor)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(TmaFenceOp)
class TmaFenceOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.tma.fence.descriptor"

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    return self.operands[0]

def tma_fence_descriptor(tensor_map_descriptor: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> TmaFenceOp:
  return TmaFenceOp(tensorMapDescriptor=tensor_map_descriptor, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class TmaPrefetchOp(_ods_ir.OpView):
  r"""
  The Op brings the cache line containing the given `$tmaDescriptor` for 
  subsequent use by the `tma.async.load` instruction.
  """

  OPERATION_NAME = "nvgpu.tma.prefetch.descriptor"

  _ODS_REGIONS = (0, True)

  def __init__(self, tensorMapDescriptor: _ods_ir.Value, *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(tensorMapDescriptor)
    if predicate is not None: operands.append(predicate)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    return None if len(self.operation.operands) < 2 else self.operation.operands[1]

@_ods_cext.register_op_adaptor(TmaPrefetchOp)
class TmaPrefetchOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.tma.prefetch.descriptor"

  @builtins.property
  def tensorMapDescriptor(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def predicate(self) -> _Optional[_ods_ir.Value[_ods_ir.IntegerType]]:
    return None if len(self.operands) < 2 else self.operands[1]

def tma_prefetch_descriptor(tensor_map_descriptor: _ods_ir.Value, *, predicate: _Optional[_ods_ir.Value[_ods_ir.IntegerType]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> TmaPrefetchOp:
  return TmaPrefetchOp(tensorMapDescriptor=tensor_map_descriptor, predicate=predicate, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class WarpgroupGenerateDescriptorOp(_ods_ir.OpView):
  r"""
  This Op builds a `nvgpu.warpgroup.descriptor` that is used by 
  `nvgpu.warpgroup.mma` to perform warpgroup-level matrix multiply and 
  accumulate.
  
  The descriptor specifies the properties of the matrix in shared memory that 
  is a multiplicand in the matrix multiply and accumulate operation. 
  """

  OPERATION_NAME = "nvgpu.warpgroup.generate.descriptor"

  _ODS_REGIONS = (0, True)

  def __init__(self, descriptor: _ods_ir.Type, tensor: _ods_ir.Value[_ods_ir.MemRefType], tensorMap: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(tensor)
    operands.append(tensorMap)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(descriptor)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def tensor(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    return self.operation.operands[0]

  @builtins.property
  def tensorMap(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def descriptor(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(WarpgroupGenerateDescriptorOp)
class WarpgroupGenerateDescriptorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.warpgroup.generate.descriptor"

  @builtins.property
  def tensor(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    return self.operands[0]

  @builtins.property
  def tensorMap(self) -> _ods_ir.Value:
    return self.operands[1]

def warpgroup_generate_descriptor(descriptor: _ods_ir.Type, tensor: _ods_ir.Value[_ods_ir.MemRefType], tensor_map: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return WarpgroupGenerateDescriptorOp(descriptor=descriptor, tensor=tensor, tensorMap=tensor_map, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class WarpgroupMmaInitAccumulatorOp(_ods_ir.OpView):
  r"""
  This Op generates and initializes the accumulator matrix for 
  `nvgpu.warpgroup.mma` op to perform matrix-multiply-and-accumulate.
  """

  OPERATION_NAME = "nvgpu.warpgroup.mma.init.accumulator"

  _ODS_REGIONS = (0, True)

  def __init__(self, matrixC: _ods_ir.Type, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(matrixC)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def matrixC(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(WarpgroupMmaInitAccumulatorOp)
class WarpgroupMmaInitAccumulatorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.warpgroup.mma.init.accumulator"

def warpgroup_mma_init_accumulator(matrix_c: _ods_ir.Type, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return WarpgroupMmaInitAccumulatorOp(matrixC=matrix_c, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class WarpgroupMmaOp(_ods_ir.OpView):
  r"""
  The `nvgpu.warpgroup.mma` op performs the warpgroup-level (4 warps) 
  matrix-multiply-and-accumulate (mma) operation that results in 
  `nvvm.wgmma.mma_async`. 
  
  The operands are `descriptorA` and `descriptorB` that are wgmma matrix 
  descriptors that shows the properties of the matrix in shared memory. The 
  results are thread-level ownership to the warpgroup-level mma operation 
  shape. The shape is deduced from the descriptor types and output vector.
  
  The Op encapsulates multiple `nvvm.wgmma.mma_async` operations to complete 
  the given shape. As `nvvm.wgmma.async` Op, or its corresponding PTX 
  instruction, is asynchronous, this Op groups the `nvvm.wgmma.async` and 
  surrounds them between `wgmma.fence.aligned` and 
  `wgmma.commit.group.sync.aligned`, `wgmma.wait.group.sync.aligned` Ops.
  
  Example:
  ```mlir
    %r1,%r2 = nvgpu.warpgroup.mma %descA, %descB, %acc1, %acc2: 
               !nvgpu.warpgroup.descriptor<tensor = memref<128x64xf16, 3>>, 
               !nvgpu.warpgroup.descriptor<tensor = memref<64x128xf16, 3>>, 
               !nvgpu.warpgroup.accumulator<fragmented = vector<64x128xf32>>,
               !nvgpu.warpgroup.accumulator<fragmented = vector<64x128xf32>>
               -> 
               !nvgpu.warpgroup.accumulator<fragmented = vector<64x128xf32>>,
               !nvgpu.warpgroup.accumulator<fragmented = vector<64x128xf32>>
  ```
  """

  OPERATION_NAME = "nvgpu.warpgroup.mma"

  _ODS_REGIONS = (0, True)

  def __init__(self, matrixD: _ods_ir.Type, descriptorA: _ods_ir.Value, descriptorB: _ods_ir.Value, matrixC: _ods_ir.Value, *, waitGroup: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, transposeA: _Optional[bool] = None, transposeB: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(descriptorA)
    operands.append(descriptorB)
    operands.append(matrixC)
    _ods_context = _ods_get_default_loc_context(loc)
    if waitGroup is not None: attributes["waitGroup"] = (waitGroup if (
        isinstance(waitGroup, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I64Attr')) else
          _ods_ir.AttrBuilder.get('I64Attr')(waitGroup, context=_ods_context))
    if bool(transposeA): attributes["transposeA"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    if bool(transposeB): attributes["transposeB"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.append(matrixD)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def descriptorA(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def descriptorB(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def matrixC(self) -> _ods_ir.Value:
    return self.operation.operands[2]

  @builtins.property
  def waitGroup(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "waitGroup" not in self.operation.attributes:
      return None
    return self.operation.attributes["waitGroup"]

  @waitGroup.setter
  def waitGroup(self, value: _Optional[_ods_ir.IntegerAttr]):
    if value is not None:
      self.operation.attributes["waitGroup"] = value
    elif "waitGroup" in self.operation.attributes:
      del self.operation.attributes["waitGroup"]

  @waitGroup.deleter
  def waitGroup(self):
    del self.operation.attributes["waitGroup"]

  @builtins.property
  def transposeA(self) -> bool:
    return "transposeA" in self.operation.attributes

  @transposeA.setter
  def transposeA(self, value):
    if bool(value):
      self.operation.attributes["transposeA"] = _ods_ir.UnitAttr.get()
    elif "transposeA" in self.operation.attributes:
      del self.operation.attributes["transposeA"]

  @transposeA.deleter
  def transposeA(self):
    del self.operation.attributes["transposeA"]

  @builtins.property
  def transposeB(self) -> bool:
    return "transposeB" in self.operation.attributes

  @transposeB.setter
  def transposeB(self, value):
    if bool(value):
      self.operation.attributes["transposeB"] = _ods_ir.UnitAttr.get()
    elif "transposeB" in self.operation.attributes:
      del self.operation.attributes["transposeB"]

  @transposeB.deleter
  def transposeB(self):
    del self.operation.attributes["transposeB"]

  @builtins.property
  def matrixD(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(WarpgroupMmaOp)
class WarpgroupMmaOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.warpgroup.mma"

  @builtins.property
  def descriptorA(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def descriptorB(self) -> _ods_ir.Value:
    return self.operands[1]

  @builtins.property
  def matrixC(self) -> _ods_ir.Value:
    return self.operands[2]

  @builtins.property
  def waitGroup(self) -> _Optional[_ods_ir.IntegerAttr]:
    if "waitGroup" not in self.attributes:
      return None
    return self.attributes["waitGroup"]

  @builtins.property
  def transposeA(self) -> bool:
    return "transposeA" in self.attributes

  @builtins.property
  def transposeB(self) -> bool:
    return "transposeB" in self.attributes

def warpgroup_mma(matrix_d: _ods_ir.Type, descriptor_a: _ods_ir.Value, descriptor_b: _ods_ir.Value, matrix_c: _ods_ir.Value, *, wait_group: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, transpose_a: _Optional[bool] = None, transpose_b: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return WarpgroupMmaOp(matrixD=matrix_d, descriptorA=descriptor_a, descriptorB=descriptor_b, matrixC=matrix_c, waitGroup=wait_group, transposeA=transpose_a, transposeB=transpose_b, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class WarpgroupMmaStoreOp(_ods_ir.OpView):
  r"""
  The `nvgpu.warpgroup.mma.store` op performs the store of fragmented result 
  in $matrixD to given memref. 
  
  [See the details of register fragment layout for accumulator matrix D]
  (https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#wgmma-64n16-d) 
  
  Note that, the op must be run with warp group.
  """

  OPERATION_NAME = "nvgpu.warpgroup.mma.store"

  _ODS_REGIONS = (0, True)

  def __init__(self, matrixD: _ods_ir.Value, dstMemref: _ods_ir.Value[_ods_ir.MemRefType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(matrixD)
    operands.append(dstMemref)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def matrixD(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def dstMemref(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    return self.operation.operands[1]

@_ods_cext.register_op_adaptor(WarpgroupMmaStoreOp)
class WarpgroupMmaStoreOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "nvgpu.warpgroup.mma.store"

  @builtins.property
  def matrixD(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def dstMemref(self) -> _ods_ir.Value[_ods_ir.MemRefType]:
    return self.operands[1]

def warpgroup_mma_store(matrix_d: _ods_ir.Value, dst_memref: _ods_ir.Value[_ods_ir.MemRefType], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> WarpgroupMmaStoreOp:
  return WarpgroupMmaStoreOp(matrixD=matrix_d, dstMemref=dst_memref, loc=loc, ip=ip)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""Internal launch-configuration normalization shared by dispatch and planning."""

_LAUNCH_CONFIG_TRACKER_METADATA_KEY = "launch_config_tracker"
_LAUNCH_CONFIG_TRACKER_OPTION = "__launch_config_tracker__"


class _LaunchConfigTracker:
    """Lazily normalize and record launch metadata used by one compiler attempt."""

    __slots__ = ("_available_launch_config", "_launch_config", "required")

    def __init__(self, available_launch_config):
        self._available_launch_config = available_launch_config
        self._launch_config = None
        self.required = False

    def require(self):
        if self._launch_config is None:
            self._launch_config = _normalize_available_launch_config(self._available_launch_config)
        self.required = True
        return self._launch_config

    @property
    def launch_config(self):
        return self._launch_config


def _is_launch_config_key_tuple(value):
    return (
        isinstance(value, tuple)
        and len(value) == 4
        and all(isinstance(item, tuple) and len(item) == 2 for item in value)
        and value[0][0] == "grid"
        and value[1][0] == "block"
        and value[2][0] == "sharedmem"
        and value[3][0] == "cluster"
    )


def _launch_config_key(launch_config):
    """Return the specialization key for a configure-produced launch config."""

    if launch_config is None:
        return None
    block = launch_config.get("block")
    if block is None:
        raise ValueError("launch_config must contain a 'block' entry")
    if not isinstance(block, tuple):
        raise TypeError("launch_config 'block' must be a normalized tuple")
    grid = launch_config.get("grid")
    if grid is None:
        raise ValueError("launch_config must contain a 'grid' entry")
    if not isinstance(grid, tuple):
        raise TypeError("launch_config 'grid' must be a normalized tuple")
    cluster = launch_config.get("cluster")
    if cluster is not None and not isinstance(cluster, tuple):
        raise TypeError("launch_config 'cluster' must be a normalized tuple or None")
    sharedmem = launch_config.get("sharedmem", 0)
    if sharedmem is None:
        sharedmem = 0
    try:
        sharedmem = int(sharedmem)
    except (TypeError, ValueError):
        raise TypeError("launch_config 'sharedmem' must be integer-convertible") from None
    return (
        ("grid", grid),
        ("block", block),
        ("sharedmem", sharedmem),
        ("cluster", cluster),
    )


def _launch_config_dict_from_key(launch_config_key):
    if not _is_launch_config_key_tuple(launch_config_key):
        raise TypeError("launch_config_key must be a normalized launch-config key")
    return {name: value for name, value in launch_config_key}


def _normalize_available_launch_config(launch_config):
    launch_config_key = _launch_config_key(launch_config)
    if launch_config_key is None:
        return None
    return _launch_config_dict_from_key(launch_config_key)



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Facility for symbolically referencing type variables.

Type variables are instances of the TypeVar class, which is uniqued by name.
An "expando" accessor `TV` is provided that generates a named TypeVar for
any attribute access:

  >>> TV.T
  TypeVar(T)
  >>> TV.T is TV.U
  False
  >>> TV.T is TV.T
  True
"""

from enum import Enum
from typing import Dict

__all__ = [
    "TypeVar",
    "TV",
    # Predefined types.
    "I32",
    "I64",
    "F32",
    "F64",
    # TypeVar aliases.
    "T",
    "U",
    "V",
]


class TypeVar:
    """A replaceable type variable.

    Type variables are uniqued by name.
    """

    ALL_TYPEVARS = dict()  # type: Dict[str, "TypeVar"]

    def __new__(cls, name: str):
        existing = cls.ALL_TYPEVARS.get(name)
        if existing is not None:
            return existing
        new = super().__new__(cls)
        new.name = name
        cls.ALL_TYPEVARS[name] = new
        return new

    def __repr__(self):
        return f"TypeVar({self.name})"

    @classmethod
    def create_expando(cls):
        """Create an expando class that creates unique type vars on attr access."""

        class ExpandoTypeVars:
            def __getattr__(self, n):
                return cls(n)

        return ExpandoTypeVars()


# Expando access via TV.foo
TV = TypeVar.create_expando()

# Predefined types.
I32 = TV.I32
I64 = TV.I64
F32 = TV.F32
F64 = TV.F64

# Some common type name aliases.
T = TV.T
U = TV.U
V = TV.V




# Autogenerated by mlir-tblgen; don't manually edit.

from enum import IntEnum, auto, IntFlag
from ._ods_common import _cext as _ods_cext
from ..ir import register_attribute_builder
_ods_ir = _ods_cext.ir

class AddressSpace(IntEnum):
    """GPU address space"""

    Global = 1
    Workgroup = 2
    Private = 3
    Constant = 4

    def __str__(self):
        if self is AddressSpace.Global:
            return "global"
        if self is AddressSpace.Workgroup:
            return "workgroup"
        if self is AddressSpace.Private:
            return "private"
        if self is AddressSpace.Constant:
            return "constant"
        raise ValueError("Unknown AddressSpace enum entry.")



@register_attribute_builder("GPU_AddressSpaceEnum", allow_existing=True)
def _gpu_addressspaceenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class AllReduceOperation(IntEnum):
    """built-in reduction operations supported by gpu.allreduce."""

    ADD = 0
    MUL = 1
    MINUI = 2
    MINSI = 3
    MINNUMF = 4
    MAXUI = 5
    MAXSI = 6
    MAXNUMF = 7
    AND = 8
    OR = 9
    XOR = 10
    MINIMUMF = 11
    MAXIMUMF = 12

    def __str__(self):
        if self is AllReduceOperation.ADD:
            return "add"
        if self is AllReduceOperation.MUL:
            return "mul"
        if self is AllReduceOperation.MINUI:
            return "minui"
        if self is AllReduceOperation.MINSI:
            return "minsi"
        if self is AllReduceOperation.MINNUMF:
            return "minnumf"
        if self is AllReduceOperation.MAXUI:
            return "maxui"
        if self is AllReduceOperation.MAXSI:
            return "maxsi"
        if self is AllReduceOperation.MAXNUMF:
            return "maxnumf"
        if self is AllReduceOperation.AND:
            return "and"
        if self is AllReduceOperation.OR:
            return "or"
        if self is AllReduceOperation.XOR:
            return "xor"
        if self is AllReduceOperation.MINIMUMF:
            return "minimumf"
        if self is AllReduceOperation.MAXIMUMF:
            return "maximumf"
        raise ValueError("Unknown AllReduceOperation enum entry.")



@register_attribute_builder("GPU_AllReduceOperation", allow_existing=True)
def _gpu_allreduceoperation(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class BroadcastType(IntEnum):
    """a lane to broadcast from"""

    first_active_lane = 0
    specific_lane = 1

    def __str__(self):
        if self is BroadcastType.first_active_lane:
            return "first_active_lane"
        if self is BroadcastType.specific_lane:
            return "specific_lane"
        raise ValueError("Unknown BroadcastType enum entry.")



@register_attribute_builder("GPU_BroadcastType", allow_existing=True)
def _gpu_broadcasttype(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class CompilationTarget(IntEnum):
    """GPU compilation format"""

    Offload = 1
    Assembly = 2
    Binary = 3
    Fatbin = 4

    def __str__(self):
        if self is CompilationTarget.Offload:
            return "offload"
        if self is CompilationTarget.Assembly:
            return "assembly"
        if self is CompilationTarget.Binary:
            return "bin"
        if self is CompilationTarget.Fatbin:
            return "fatbin"
        raise ValueError("Unknown CompilationTarget enum entry.")



@register_attribute_builder("GPU_CompilationTargetEnum", allow_existing=True)
def _gpu_compilationtargetenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class Dimension(IntEnum):
    """a dimension, either 'x', 'y', or 'z'"""

    x = 0
    y = 1
    z = 2

    def __str__(self):
        if self is Dimension.x:
            return "x"
        if self is Dimension.y:
            return "y"
        if self is Dimension.z:
            return "z"
        raise ValueError("Unknown Dimension enum entry.")



@register_attribute_builder("GPU_Dimension", allow_existing=True)
def _gpu_dimension(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class DimensionKind(IntEnum):
    """the possible kinds of launch dimension"""

    Other = 0
    Block = 1
    Grid = 2
    Cluster = 3

    def __str__(self):
        if self is DimensionKind.Other:
            return "other"
        if self is DimensionKind.Block:
            return "block"
        if self is DimensionKind.Grid:
            return "grid"
        if self is DimensionKind.Cluster:
            return "cluster"
        raise ValueError("Unknown DimensionKind enum entry.")



class Prune2To4SpMatFlag(IntEnum):
    """pruning strategy for 2:4 sparse matrix"""

    NONE = 0
    PRUNE_ONLY = 1
    PRUNE_AND_CHECK = 2

    def __str__(self):
        if self is Prune2To4SpMatFlag.NONE:
            return "NONE"
        if self is Prune2To4SpMatFlag.PRUNE_ONLY:
            return "PRUNE_ONLY"
        if self is Prune2To4SpMatFlag.PRUNE_AND_CHECK:
            return "PRUNE_AND_CHECK"
        raise ValueError("Unknown Prune2To4SpMatFlag enum entry.")



@register_attribute_builder("GPU_Prune2To4SpMatFlag", allow_existing=True)
def _gpu_prune2to4spmatflag(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class ShuffleMode(IntEnum):
    """Indexing modes supported by gpu.shuffle."""

    XOR = 0
    UP = 2
    DOWN = 1
    IDX = 3

    def __str__(self):
        if self is ShuffleMode.XOR:
            return "xor"
        if self is ShuffleMode.UP:
            return "up"
        if self is ShuffleMode.DOWN:
            return "down"
        if self is ShuffleMode.IDX:
            return "idx"
        raise ValueError("Unknown ShuffleMode enum entry.")



@register_attribute_builder("GPU_ShuffleMode", allow_existing=True)
def _gpu_shufflemode(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class SpGEMMWorkEstimationOrComputeKind(IntEnum):
    """choose whether spgemm_work_estimation_or_compute does work estimation or compute"""

    WORK_ESTIMATION = 0
    COMPUTE = 1

    def __str__(self):
        if self is SpGEMMWorkEstimationOrComputeKind.WORK_ESTIMATION:
            return "WORK_ESTIMATION"
        if self is SpGEMMWorkEstimationOrComputeKind.COMPUTE:
            return "COMPUTE"
        raise ValueError("Unknown SpGEMMWorkEstimationOrComputeKind enum entry.")



@register_attribute_builder("GPU_SpGEMMWorkEstimationOrComputeKind", allow_existing=True)
def _gpu_spgemmworkestimationorcomputekind(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class TransposeMode(IntEnum):
    """transpose mode of sparse matrix supported by sparse tensor ops"""

    NON_TRANSPOSE = 0
    TRANSPOSE = 1
    CONJUGATE_TRANSPOSE = 2

    def __str__(self):
        if self is TransposeMode.NON_TRANSPOSE:
            return "NON_TRANSPOSE"
        if self is TransposeMode.TRANSPOSE:
            return "TRANSPOSE"
        if self is TransposeMode.CONJUGATE_TRANSPOSE:
            return "CONJUGATE_TRANSPOSE"
        raise ValueError("Unknown TransposeMode enum entry.")



@register_attribute_builder("GPU_TransposeMode", allow_existing=True)
def _gpu_transposemode(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class MMAElementwiseOp(IntEnum):
    """elementwise operation to apply to mma matrix"""

    ADDF = 0
    MULF = 1
    SUBF = 2
    MAXF = 3
    MINF = 4
    DIVF = 5
    ADDI = 6
    MULI = 7
    SUBI = 8
    DIVS = 9
    DIVU = 10
    NEGATEF = 11
    NEGATES = 12
    EXTF = 13
    TRUNCF = 14

    def __str__(self):
        if self is MMAElementwiseOp.ADDF:
            return "addf"
        if self is MMAElementwiseOp.MULF:
            return "mulf"
        if self is MMAElementwiseOp.SUBF:
            return "subf"
        if self is MMAElementwiseOp.MAXF:
            return "maxf"
        if self is MMAElementwiseOp.MINF:
            return "minf"
        if self is MMAElementwiseOp.DIVF:
            return "divf"
        if self is MMAElementwiseOp.ADDI:
            return "addi"
        if self is MMAElementwiseOp.MULI:
            return "muli"
        if self is MMAElementwiseOp.SUBI:
            return "subi"
        if self is MMAElementwiseOp.DIVS:
            return "divs"
        if self is MMAElementwiseOp.DIVU:
            return "divu"
        if self is MMAElementwiseOp.NEGATEF:
            return "negatef"
        if self is MMAElementwiseOp.NEGATES:
            return "negates"
        if self is MMAElementwiseOp.EXTF:
            return "extf"
        if self is MMAElementwiseOp.TRUNCF:
            return "truncf"
        raise ValueError("Unknown MMAElementwiseOp enum entry.")



@register_attribute_builder("MMAElementWise", allow_existing=True)
def _mmaelementwise(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class MappingId(IntEnum):
    """Mapping ids for loop mapping"""

    DimX = 0
    DimY = 1
    DimZ = 2
    LinearDim0 = 3
    LinearDim1 = 4
    LinearDim2 = 5
    LinearDim3 = 6
    LinearDim4 = 7
    LinearDim5 = 8
    LinearDim6 = 9
    LinearDim7 = 10
    LinearDim8 = 11
    LinearDim9 = 12

    def __str__(self):
        if self is MappingId.DimX:
            return "x"
        if self is MappingId.DimY:
            return "y"
        if self is MappingId.DimZ:
            return "z"
        if self is MappingId.LinearDim0:
            return "linear_dim_0"
        if self is MappingId.LinearDim1:
            return "linear_dim_1"
        if self is MappingId.LinearDim2:
            return "linear_dim_2"
        if self is MappingId.LinearDim3:
            return "linear_dim_3"
        if self is MappingId.LinearDim4:
            return "linear_dim_4"
        if self is MappingId.LinearDim5:
            return "linear_dim_5"
        if self is MappingId.LinearDim6:
            return "linear_dim_6"
        if self is MappingId.LinearDim7:
            return "linear_dim_7"
        if self is MappingId.LinearDim8:
            return "linear_dim_8"
        if self is MappingId.LinearDim9:
            return "linear_dim_9"
        raise ValueError("Unknown MappingId enum entry.")



@register_attribute_builder("MappingIdEnum", allow_existing=True)
def _mappingidenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

class Processor(IntEnum):
    """processor for loop mapping"""

    BlockX = 0
    BlockY = 1
    BlockZ = 2
    ThreadX = 3
    ThreadY = 4
    ThreadZ = 5
    Sequential = 6

    def __str__(self):
        if self is Processor.BlockX:
            return "block_x"
        if self is Processor.BlockY:
            return "block_y"
        if self is Processor.BlockZ:
            return "block_z"
        if self is Processor.ThreadX:
            return "thread_x"
        if self is Processor.ThreadY:
            return "thread_y"
        if self is Processor.ThreadZ:
            return "thread_z"
        if self is Processor.Sequential:
            return "sequential"
        raise ValueError("Unknown Processor enum entry.")



@register_attribute_builder("ProcessorEnum", allow_existing=True)
def _processorenum(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(64, context=context), int(x))

@register_attribute_builder("gpu.GPU_AddressSpaceAttr")
def _gpu_addressspaceattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu.address_space<{str(x)}>', context=context)

@register_attribute_builder("gpu.GPU_AllReduceOperationAttr")
def _gpu_allreduceoperationattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<all_reduce_op {str(x)}>', context=context)

@register_attribute_builder("gpu.GPU_BroadcastTypeAttr")
def _gpu_broadcasttypeattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<broadcast {str(x)}>', context=context)

@register_attribute_builder("gpu.GPU_DimensionAttr")
def _gpu_dimensionattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<dim {str(x)}>', context=context)

@register_attribute_builder("gpu.GPU_Prune2To4SpMatFlagAttr")
def _gpu_prune2to4spmatflagattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<prune_2to4_spmat_flag {str(x)}>', context=context)

@register_attribute_builder("gpu.GPU_ShuffleModeAttr")
def _gpu_shufflemodeattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<shuffle_mode {str(x)}>', context=context)

@register_attribute_builder("gpu.GPU_SpGEMMWorkEstimationOrComputeKindAttr")
def _gpu_spgemmworkestimationorcomputekindattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<spgemm_work_estimation_or_compute_kind {str(x)}>', context=context)

@register_attribute_builder("gpu.GPU_TransposeModeAttr")
def _gpu_transposemodeattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<mat_transpose_mode {str(x)}>', context=context)

@register_attribute_builder("gpu.MMAElementWiseAttr")
def _mmaelementwiseattr(x, context):
    return _ods_ir.Attribute.parse(f'#gpu<mma_element_wise {str(x)}>', context=context)




# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
import operator
from abc import abstractmethod
from copy import deepcopy
from functools import cached_property, partialmethod
from typing import Optional, Tuple, Union

import numpy as np

from ..util import infer_mlir_type, mlir_type_to_np_dtype
from numba_cuda_mlir._mlir._mlir_libs._mlir import register_value_caster
from numba_cuda_mlir._mlir.dialects import complex as complex_dialect
from numba_cuda_mlir._mlir.dialects import math as math_dialect
from numba_cuda_mlir._mlir.dialects._arith_enum_gen import (
    _arith_cmpfpredicateattr,
    _arith_cmpipredicateattr,
)
from numba_cuda_mlir._mlir.dialects._ods_common import (
    get_op_result_or_op_results,
    get_op_result_or_value,
)
from numba_cuda_mlir._mlir.dialects.arith import *
from numba_cuda_mlir._mlir.ir import (
    Attribute,
    BF16Type,
    ComplexType,
    Context,
    DenseElementsAttr,
    F16Type,
    F32Type,
    F64Type,
    FloatAttr,
    FloatType,
    IndexType,
    InsertionPoint,
    IntegerType,
    Location,
    OpView,
    Operation,
    ShapedType,
    Type,
    Value,
    register_attribute_builder,
)


def constant(
    value: Union[int, float, bool, np.ndarray],
    type: Optional[Type] = None,
    index: Optional[bool] = None,
    *,
    result: Optional[Type] = None,
    vector: Optional[bool] = False,
    loc: Location = None,
    ip: InsertionPoint = None,
) -> Value:
    """Instantiate arith.constant with value `value`.

    Args:
      value: Python value that determines the value attribute of the
        arith.constant op.
      type: Optional MLIR type that type of the value attribute of the
        arith.constant op; if omitted the type of the value attribute
        will be inferred from the value.
      index: Whether the MLIR type should be an index type; if passed the
        type argument will be ignored.

    Returns:
      ir.OpView instance that corresponds to instantiated arith.constant op.
    """
    if index is not None and index:
        type = IndexType.get()

    if type is not None and result is not None:
        raise ValueError("only type XOR result expected")
    type = type or result
    if type is None:
        type = infer_mlir_type(value, vector=vector)

    assert type is not None

    if isinstance(type, ComplexType):
        value = complex(value)
        return get_op_result_or_op_results(
            complex_dialect.ConstantOp(
                type,
                list(
                    map(
                        lambda x: FloatAttr.get(type.element_type, x),
                        [value.real, value.imag],
                    )
                ),
                loc=loc,
                ip=ip,
            )
        )

    if isinstance(type, FloatType) and not isinstance(value, np.ndarray):
        value = float(value)

    if isinstance(type, ShapedType) and isinstance(value, (int, float, bool)):
        ranked_tensor_type = ShapedType(type)
        value = np.full(
            ranked_tensor_type.shape,
            value,
            dtype=mlir_type_to_np_dtype(ranked_tensor_type.element_type),
        )

    if isinstance(value, np.ndarray):
        value = DenseElementsAttr.get(
            value,
            type=type,
        )

    return get_op_result_or_op_results(ConstantOp(type, value, loc=loc, ip=ip))


def index_cast(
    in_: Value,
    to: Type = None,
    *,
    out: Type = None,
    loc: Location = None,
    ip: InsertionPoint = None,
) -> Value:
    assert bool(to) != bool(out), "either `to` or `out` must be set but not both"
    res_type = out or to
    if res_type is None:
        res_type = IndexType.get()
    return get_op_result_or_op_results(IndexCastOp(res_type, in_, loc=loc, ip=ip))


@register_attribute_builder("Arith_CmpIPredicateAttr", replace=True)
def _arith_CmpIPredicateAttr(predicate: Union[str, Attribute], context: Context):
    predicates = {
        "eq": CmpIPredicate.eq,
        "ne": CmpIPredicate.ne,
        "slt": CmpIPredicate.slt,
        "sle": CmpIPredicate.sle,
        "sgt": CmpIPredicate.sgt,
        "sge": CmpIPredicate.sge,
        "ult": CmpIPredicate.ult,
        "ule": CmpIPredicate.ule,
        "ugt": CmpIPredicate.ugt,
        "uge": CmpIPredicate.uge,
        0: CmpIPredicate.eq,
        1: CmpIPredicate.ne,
        2: CmpIPredicate.slt,
        3: CmpIPredicate.sle,
        4: CmpIPredicate.sgt,
        5: CmpIPredicate.sge,
        6: CmpIPredicate.ult,
        7: CmpIPredicate.ule,
        8: CmpIPredicate.ugt,
        9: CmpIPredicate.uge,
    }
    if isinstance(predicate, Attribute):
        return predicate
    predicate = str(predicate)
    assert predicate in predicates, f"{predicate=} not in predicates"
    return _arith_cmpipredicateattr(predicates[predicate], context)


@register_attribute_builder("Arith_CmpFPredicateAttr", replace=True)
def _arith_CmpFPredicateAttr(predicate: Union[str, Attribute], context: Context):
    predicates = {
        "false": CmpFPredicate.AlwaysFalse,
        # ordered comparison
        # An ordered comparison checks if neither operand is NaN.
        "oeq": CmpFPredicate.OEQ,
        "ogt": CmpFPredicate.OGT,
        "oge": CmpFPredicate.OGE,
        "olt": CmpFPredicate.OLT,
        "ole": CmpFPredicate.OLE,
        "one": CmpFPredicate.ONE,
        # no clue what this one is
        "ord": CmpFPredicate.ORD,
        # unordered comparison
        # Conversely, an unordered comparison checks if either operand is a NaN.
        "ueq": CmpFPredicate.UEQ,
        "ugt": CmpFPredicate.UGT,
        "uge": CmpFPredicate.UGE,
        "ult": CmpFPredicate.ULT,
        "ule": CmpFPredicate.ULE,
        "une": CmpFPredicate.UNE,
        # no clue what this one is
        "uno": CmpFPredicate.UNO,
        # return always true
        "true": CmpFPredicate.AlwaysTrue,
    }
    if isinstance(predicate, Attribute):
        return predicate
    predicate = str(predicate)
    assert predicate in predicates, f"{predicate=} not in predicates"
    return _arith_cmpfpredicateattr(predicates[predicate], context)


def _binary_op(
    lhs: "ArithValue",
    rhs: "ArithValue",
    op: str,
    predicate: str = None,
    signedness: str = None,
    *,
    loc: Location = None,
) -> "ArithValue":
    """Generic for handling infix binary operator dispatch.

    Args:
      lhs: E.g. ScalarValue or TensorValue below.
      rhs: ScalarValue or TensorValue with type matching self.
      op: Binary operator, currently only add, sub, mul
        supported.

    Returns:
      Result of binary operation. This will be a handle to an arith(add|sub|mul) op.
    """
    if (isinstance(lhs, Value) and isinstance(rhs, Value) and lhs.type != rhs.type) or isinstance(
        rhs, (float, int, bool, np.ndarray)
    ):
        rhs = lhs.coerce(rhs)
    elif isinstance(rhs, Value) and isinstance(lhs, (float, int, bool, np.ndarray)):
        lhs = rhs.coerce(lhs)

    assert lhs.type == rhs.type, f"{lhs=} {rhs=} must have the same type."

    assert op in {
        "add",
        "and",
        "or",
        "xor",
        "sub",
        "mul",
        "cmp",
        "truediv",
        "floordiv",
        "mod",
    }

    if op == "cmp":
        assert predicate is not None

    if lhs.fold() and lhs.fold():
        klass = lhs.__class__
        # if both operands are constants (results of an arith.constant op)
        # then both have a literal value (i.e. Python value).
        lhs, rhs = lhs.literal_value, rhs.literal_value
        # if we're folding constants (self._fold = True) then we just carry out
        # the corresponding operation on the literal values; e.g., operator.add.
        # note this is the same as op = operator.__dict__[op].
        if predicate is not None:
            op = predicate
        op = operator.attrgetter(op)(operator)
        return klass(op(lhs, rhs), fold=True)

    if op == "truediv":
        op = "div"
    if op == "mod":
        op = "rem"

    op = op.capitalize()
    if isinstance(lhs.dtype, FloatType):
        if op == "Xor":
            raise TypeError(
                f"unsupported operand type(s) for ^: '{type(lhs).__name__}' and "
                f"'{type(rhs).__name__}'"
            )
        if op == "Floordiv":
            # Python float // float is floor(a / b); arith has no floordivf — match
            # numba_cuda_mlir/lowering/math.py floordiv_cg: divf then math.floor.
            div_op = globals()["DivFOp"]
            div_val = lhs.__class__(div_op(lhs, rhs, loc=loc), dtype=lhs.dtype)
            return lhs.__class__(math_dialect.floor(div_val, loc=loc), dtype=lhs.dtype)
        op += "F"
    elif isinstance(lhs.dtype, (IntegerType, IndexType)):
        if "div" in op.lower() or "rem" in op.lower():
            if not isinstance(lhs.dtype, IndexType) and not lhs.dtype.is_signless:
                raise ValueError(f"{op.lower()}i not supported for {lhs=}")
            if op == "Floordiv":
                op = "FloorDiv"
            op += "S"
        op += "I"
    else:
        raise NotImplementedError(f"Unsupported '{op}' operands: {lhs}, {rhs}")

    # because up above we have from arith import *
    # MLIR ODS name is XOrIOp (capital O), not XorIOp, after "xor" -> capitalize -> "Xor" + "I".
    op_cls = f"{op}Op"
    if op_cls == "XorIOp":
        op_cls = "XOrIOp"
    op = globals()[op_cls]

    if predicate is not None:
        if isinstance(lhs.dtype, FloatType):
            # ordered comparison - see above
            predicate = "o" + predicate
        elif isinstance(lhs.dtype, (IntegerType, IndexType)):
            # eq, ne signs don't matter
            if predicate not in {"eq", "ne"}:
                if signedness is not None:
                    predicate = signedness + predicate
                else:
                    if isinstance(lhs.dtype, IndexType) or lhs.dtype.is_unsigned:
                        predicate = "u" + predicate
                    else:
                        assert lhs.dtype.is_signed or lhs.dtype.is_signless
                        predicate = "s" + predicate
        return lhs.__class__(op(predicate, lhs, rhs, loc=loc), dtype=lhs.dtype)
    else:
        return lhs.__class__(op(lhs, rhs, loc=loc), dtype=lhs.dtype)


def _rbinary_op(
    rhs: "ArithValue",
    lhs: "ArithValue",
    op: str,
    predicate: str = None,
    signedness: str = None,
    *,
    loc: Location = None,
) -> "ArithValue":
    return _binary_op(lhs, rhs, op, predicate, signedness, loc=loc)


def _pow_op(
    lhs: "ArithValue",
    rhs: "ArithValue",
    *,
    loc: Location = None,
) -> "ArithValue":
    if isinstance(rhs, (float, int, bool)):
        rhs = lhs.coerce(rhs)
    elif isinstance(lhs, (float, int, bool)):
        lhs = rhs.coerce(lhs)

    if lhs.fold() and rhs.fold():
        klass = lhs.__class__
        return klass(operator.pow(lhs.literal_value, rhs.literal_value), fold=True)

    if isinstance(lhs.dtype, FloatType) and isinstance(rhs.dtype, FloatType):
        return lhs.__class__(math_dialect.powf(lhs, rhs, loc=loc), dtype=lhs.dtype)
    elif isinstance(lhs.dtype, FloatType) and isinstance(rhs.dtype, IntegerType):
        return lhs.__class__(math_dialect.fpowi(lhs, rhs, loc=loc), dtype=lhs.dtype)
    elif isinstance(lhs.dtype, IntegerType) and isinstance(rhs.dtype, IntegerType):
        assert lhs.type == rhs.type, f"{lhs=} {rhs=} must have the same type."
        from numba_cuda_mlir.mlir.dialect_exts.math import ipowi

        return lhs.__class__(ipowi(lhs, rhs, loc=loc), dtype=lhs.dtype)
    else:
        raise NotImplementedError(f"Unsupported pow operands: {lhs.type}, {rhs.type}")


def _rpow_op(
    rhs: "ArithValue",
    lhs: "ArithValue",
    *,
    loc: Location = None,
) -> "ArithValue":
    return _pow_op(lhs, rhs, loc=loc)


class ArithValue(Value, metaclass=type(Value)):
    """Class for functionality shared by Value subclasses that support
    arithmetic operations.

    Wrapping logic is implemented in `__new__` so that nanobind's own
    instance-creation path (`nb_type.__call__`) is used. That path knows how
    to initialize a Python subclass of a nanobind-bound `Value` correctly;
    calling `cls.__init__` explicitly from a Python metaclass does not.

    The class uses `type(Value)` (nanobind's `nb_type`) as its metaclass.
    This lets nanobind recognise `ArithValue` and its subclasses as the
    Python wrapper for MLIR `Value` objects, so block arguments are
    automatically wrapped, while avoiding the metaclass-subclass check that
    nanobind 2.13 tightened.
    """

    def __new__(cls, arg, *, dtype=None, fold=None, loc=None, ip=None):
        if isinstance(arg, (OpView, Operation, Value)):
            if isinstance(arg, (Operation, OpView)):
                assert len(arg.results) == 1
            val = get_op_result_or_value(arg)
        elif isinstance(arg, (int, float, bool, np.ndarray)):
            if dtype is not None and not isinstance(dtype, Type):
                raise ValueError(f"{dtype=} is expected to be an ir.Type.")
            if fold is not None and not isinstance(fold, bool):
                raise ValueError(f"{fold=} is expected to be a bool.")
            val = constant(deepcopy(arg), dtype, loc=loc, ip=ip)
        else:
            raise NotImplementedError(f"{cls.__name__} doesn't support wrapping {arg}.")
        instance = Value.__new__(cls, val)
        instance._arith_value = val
        return instance

    def __init__(self, arg, *, dtype=None, fold: Optional[bool] = None, loc=None, ip=None):
        super().__init__(self._arith_value)
        del self._arith_value
        self._fold = fold if fold is not None else False

    def is_constant(self) -> bool:
        return isinstance(self.owner, OpView) and isinstance(self.owner, ConstantOp)

    @property
    @abstractmethod
    def literal_value(self):
        pass

    @abstractmethod
    def coerce(self, other) -> Tuple["ArithValue", "ArithValue"]:
        pass

    def __hash__(self):
        return Value(self).__hash__()

    def fold(self) -> bool:
        return self.is_constant() and self._fold

    def __str__(self):
        return f"{self.__class__.__name__}({self.get_name()}, {self.type})"

    def __repr__(self):
        return str(Value(self)).replace("Value", self.__class__.__name__)

    # partialmethod differs from partial in that it also binds the object instance
    # to the first arg (i.e., self)
    __add__ = partialmethod(_binary_op, op="add")
    __sub__ = partialmethod(_binary_op, op="sub")
    __mul__ = partialmethod(_binary_op, op="mul")
    __truediv__ = partialmethod(_binary_op, op="truediv")
    __floordiv__ = partialmethod(_binary_op, op="floordiv")
    __mod__ = partialmethod(_binary_op, op="mod")
    __and__ = partialmethod(_binary_op, op="and")
    __or__ = partialmethod(_binary_op, op="or")
    __xor__ = partialmethod(_binary_op, op="xor")
    __pow__ = partialmethod(_pow_op)

    __radd__ = partialmethod(_rbinary_op, op="add")
    __rsub__ = partialmethod(_rbinary_op, op="sub")
    __rmul__ = partialmethod(_rbinary_op, op="mul")
    __rtruediv__ = partialmethod(_rbinary_op, op="truediv")
    __rfloordiv__ = partialmethod(_rbinary_op, op="floordiv")
    __rmod__ = partialmethod(_rbinary_op, op="mod")
    __rand__ = partialmethod(_rbinary_op, op="and")
    __ror__ = partialmethod(_rbinary_op, op="or")
    __rxor__ = partialmethod(_rbinary_op, op="xor")
    __rpow__ = partialmethod(_rpow_op)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            try:
                other = self.__class__(other, dtype=self.type)
            except NotImplementedError as e:
                assert "doesn't support wrapping" in str(e)
                return False
        if self is other:
            return True
        return _binary_op(self, other, op="cmp", predicate="eq")

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            try:
                other = self.__class__(other, dtype=self.type)
            except NotImplementedError as e:
                assert "doesn't support wrapping" in str(e)
                return True
        if self is other:
            return False
        return _binary_op(self, other, op="cmp", predicate="ne")

    __le__ = partialmethod(_binary_op, op="cmp", predicate="le")
    __lt__ = partialmethod(_binary_op, op="cmp", predicate="lt")
    __ge__ = partialmethod(_binary_op, op="cmp", predicate="ge")
    __gt__ = partialmethod(_binary_op, op="cmp", predicate="gt")

    __rle__ = partialmethod(_rbinary_op, op="cmp", predicate="le")
    __rlt__ = partialmethod(_rbinary_op, op="cmp", predicate="lt")
    __rge__ = partialmethod(_rbinary_op, op="cmp", predicate="ge")
    __rgt__ = partialmethod(_rbinary_op, op="cmp", predicate="gt")

    def _eq(self, other):
        return Value(self) == Value(other)

    def _ne(self, other):
        return Value(self) != Value(other)


class ScalarValue(ArithValue):
    """Value subclass ScalarValue that adds convenience methods
    for getting dtype and (possibly) the stored literal value.

    Note, order matters in the superclasses above; ArithValue is first so that
    e.g. __init__, and __str__ from ArithValue are used instead of
    from ScalarValue.
    """

    @cached_property
    def dtype(self) -> Type:
        return self.type

    @cached_property
    def literal_value(self) -> Union[int, float, bool]:
        if not self.is_constant():
            raise ValueError("Can't build literal from non-constant ScalarValue")
        return self.owner.opview.literal_value

    def __int__(self):
        return int(self.literal_value)

    def __float__(self):
        return float(self.literal_value)

    def coerce(self, other) -> Tuple["ScalarValue", "ScalarValue"]:
        if isinstance(other, (int, float, bool)):
            other = ScalarValue(other, dtype=self.dtype)
        elif isinstance(other, ScalarValue) and (
            isinstance(self.type, IndexType) or isinstance(other.type, IndexType)
        ):
            other = index_cast(other, to=self.type)
        else:
            raise ValueError(f"can't coerce {other=} to {self=}")
        return other


for t in [BF16Type, F16Type, F32Type, F64Type, IndexType, IntegerType, ComplexType]:
    register_value_caster(t.static_typeid)(ScalarValue)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from numba_cuda_mlir.extending import as_numba_type, overload, typing_registry
from numba_cuda_mlir.numba_cuda.core.errors import NumbaTypeError
from numba_cuda_mlir import types
import functools
import operator

overload = functools.partial(overload, typing_registry=typing_registry)


@overload(operator.truth)
def ol_truth(val):
    if isinstance(val, types.Boolean):

        def impl(val):
            return val

        return impl


@overload(bool)
def bool_sequence(x):
    valid_types = (
        types.CharSeq,
        types.UnicodeCharSeq,
        types.UnicodeType,
    )

    if isinstance(x, valid_types):

        def bool_impl(x):
            return len(x) > 0

        return bool_impl


@overload(bool, inline="always")
def bool_none(x):
    if isinstance(x, types.NoneType) or x is None:
        return lambda x: False


@overload(str)
def ol_str_generic(object=""):
    def impl(object=""):
        attr = "__str__"
        if hasattr(object, attr):
            return getattr(object, attr)()
        else:
            return repr(object)

    return impl


@overload(isinstance)
def ol_isinstance(var, typs):
    def true_impl(var, typs):
        return True

    def false_impl(var, typs):
        return False

    var_ty = as_numba_type(var)

    if isinstance(var_ty, types.Optional):
        msg = f'isinstance cannot handle optional types. Found: "{var_ty}"'
        raise NumbaTypeError(msg)

    # NOTE: The current implementation of `isinstance` restricts the type of the
    # instance variable to types that are well known and in common use. The
    # danger of unrestricted type comparison is that a "default" of `False` is
    # required and this means that if there is a bug in the logic of the
    # comparison tree `isinstance` returns False! It's therefore safer to just
    # reject the compilation as untypable!
    supported_var_ty = (
        types.Number,
        types.Bytes,
        types.RangeType,
        types.Tuple,
        types.UniTuple,
        types.Function,
        types.UnicodeType,
        types.NoneType,
        types.Array,
        types.Boolean,
        types.Float,
        types.UnicodeCharSeq,
        types.Complex,
        types.NPDatetime,
        types.NPTimedelta,
    )
    if not isinstance(var_ty, supported_var_ty):
        msg = f'isinstance() does not support variables of type "{var_ty}".'
        raise NumbaTypeError(msg)

    t_typs = typs

    # Check the types that the var can be an instance of, it'll be a scalar,
    # a unituple or a tuple.
    if isinstance(t_typs, types.UniTuple):
        # corner case - all types in isinstance are the same
        t_typs = t_typs.key[0]

    if not isinstance(t_typs, types.Tuple):
        t_typs = (t_typs,)

    for typ in t_typs:
        if isinstance(typ, types.Function):
            key = typ.key[0]  # functions like int(..), float(..), str(..)
        else:
            key = typ.key

        # corner cases for bytes, range, ...
        # avoid registering those types on `as_numba_type`
        types_not_registered = {
            bytes: types.Bytes,
            range: types.RangeType,
            tuple: types.BaseTuple,
        }
        if key in types_not_registered:
            if isinstance(var_ty, types_not_registered[key]):
                return true_impl
            continue

        if isinstance(typ, types.TypeRef):
            # Use of Numba type classes is in general not supported as they do
            # not work when the jit is disabled.
            if key not in (types.ListType, types.DictType):
                msg = "Numba type classes (except numba.typed.* container types) are not supported."
                raise NumbaTypeError(msg)
            # Case for TypeRef (i.e. isinstance(var, typed.List))
            #      var_ty == ListType[int64] (instance)
            #         typ == types.ListType  (class)
            return true_impl if type(var_ty) is key else false_impl
        else:
            numba_typ = as_numba_type(key)
            if var_ty == numba_typ:
                return true_impl
            elif isinstance(numba_typ, (types.NPDatetime, types.NPTimedelta)):
                if isinstance(var_ty, type(numba_typ)):
                    return true_impl
            elif (
                isinstance(numba_typ, types.ClassType)
                and isinstance(var_ty, types.ClassInstanceType)
                and var_ty.key == numba_typ.instance_type.key
            ):
                # check for jitclasses
                return true_impl
            elif isinstance(numba_typ, types.Container) and numba_typ.key[0] == types.undefined:
                # check for containers (list, tuple, set, ...)
                if isinstance(var_ty, numba_typ.__class__) or (
                    isinstance(var_ty, types.BaseTuple) and isinstance(numba_typ, types.BaseTuple)
                ):
                    return true_impl

    return false_impl



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import sys
from typing import Any, Callable, ParamSpec, TypeVar

if sys.version_info >= (3, 11):
    from typing import TypeVarTuple, Unpack
else:
    from typing_extensions import TypeVarTuple, Unpack
from numba_cuda_mlir.numba_cuda.core import ir

# Prototype for intrinsic code generation functions
Builder = Callable[["MLIRLower", ir.Var, list[ir.Var], list[tuple[str, ir.Var]]], None]

PS = ParamSpec("PS")
AnyCallable = Callable[PS, Any]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import threading


class _LockedCounter:
    def __init__(self, start=0):
        self._value = start
        self._lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self._lock:
            value = self._value
            self._value += 1
            return value
